//
//  ViewController.m
//  IsObjcObject
//
//  Created by Alexandre Colucci on 19.11.2016.
//  Copyright © 2016 Alexandre Colucci. All rights reserved.
//

#import "ViewController.h"
#include "IsObjcObject.h"


typedef enum : NSUInteger
{
    kCheckPointerResult_NotAnObject,
    kCheckPointerResult_TaggedPointed,
    kCheckPointerResult_Object,
} CheckPointerResult;


@implementation ViewController


- (void)viewDidLoad
{
	[super viewDidLoad];
	
	[self runTests];
}

#pragma  mark - Tests

- (void)runTests
{
	NSString *startLog = [NSString stringWithFormat:@"Start time %@\n\n", [NSDateFormatter localizedStringFromDate:[NSDate date] dateStyle:NSDateFormatterShortStyle timeStyle:NSDateFormatterMediumStyle]];
	[self logText:startLog isError:YES];
	
	// NSObject
	NSObject *myObj = [[NSObject alloc] init];
	[self testPointer:(__bridge const void *)(myObj) shouldBeOfType:kCheckPointerResult_Object];
	
	// NSBundle
	NSBundle *myBundle = [NSBundle mainBundle];
	[self testPointer:(__bridge const void *)(myBundle) shouldBeOfType:kCheckPointerResult_Object printObjectDescription:NO];
	
	// infoDictionary and all objects
	NSDictionary<NSString *, id> *infoDictionary = [myBundle infoDictionary];
	[self testPointer:(__bridge const void *)(infoDictionary) shouldBeOfType:kCheckPointerResult_Object printObjectDescription:NO];
	for(NSString *infoString in infoDictionary)
	{
		[self testPointer:(__bridge const void *)(infoString) shouldNotBeOfType:kCheckPointerResult_NotAnObject];
	}
	
	// NSString
	NSString *myString = @"Long string http://blog.timac.org/?p=1454";
	[self testPointer:(__bridge const void *)(myString) shouldBeOfType:kCheckPointerResult_Object];
	
	// Various invalid addresses
	[self testFromPointer:0x0 toPtr:0x08 shouldBeOfType:kCheckPointerResult_NotAnObject];
	[self testFromPointer:0x1000 toPtr:0x1008 shouldBeOfType:kCheckPointerResult_NotAnObject];
	[self testFromPointer:0x10000 toPtr:0x10008 shouldBeOfType:kCheckPointerResult_NotAnObject];
	[self testFromPointer:0x100000 toPtr:0x100008 shouldBeOfType:kCheckPointerResult_NotAnObject];
	[self testFromPointer:0x1000000 toPtr:0x1000008 shouldBeOfType:kCheckPointerResult_NotAnObject];
	[self testFromPointer:0x0000000400000000 toPtr:0x0000000400000008 shouldBeOfType:kCheckPointerResult_NotAnObject];
	
	
	// NSString Tagged pointer
	[self testPointer:(const void *)0xa65646f635854447 shouldBeOfType:kCheckPointerResult_TaggedPointed];
	
	// NSNumber Tagged pointer
	[self testPointer:(const void *)0xb000000000004d22 shouldBeOfType:kCheckPointerResult_TaggedPointed];
	
	// Other tagged pointers
	[self testFromPointer:0x8000000000000000 toPtr:0x8000000000000008 shouldBeOfType:kCheckPointerResult_TaggedPointed];

	// Tests Completed
	NSString *completeLog = [NSString stringWithFormat:@"\n\nCompleted at %@\n", [NSDateFormatter localizedStringFromDate:[NSDate date] dateStyle:NSDateFormatterShortStyle timeStyle:NSDateFormatterMediumStyle]];
	[self logText:completeLog isError:YES];
}


-(void)testPointer:(const void *)inPtr shouldBeOfType:(CheckPointerResult)inShouldBeOfType
{
	[self testPointer:inPtr shouldBeOfType:inShouldBeOfType printObjectDescription:YES];
}

-(void)testFromPointer:(uintptr_t)fromPtr toPtr:(uintptr_t)toPtr shouldBeOfType:(CheckPointerResult)inShouldBeOfType
{
	for(uintptr_t ptr = fromPtr ; ptr < toPtr ; ptr++)
	{
		[self testPointer:(const void *)ptr shouldBeOfType:inShouldBeOfType];
	}
}

-(void)testPointer:(const void *)inPtr shouldBeOfType:(CheckPointerResult)inShouldBeOfType printObjectDescription:(BOOL)inPrintObjectDescription
{
	if([self checkPointer:inPtr printObjectDescription:inPrintObjectDescription] != inShouldBeOfType)
	{
		[self logText:@"*** Test failed" isError:YES];
	}
}

-(void)testPointer:(const void *)inPtr shouldNotBeOfType:(CheckPointerResult)inNotType
{
	if([self checkPointer:inPtr printObjectDescription:YES] == inNotType)
	{
		[self logText:@"*** Test failed" isError:YES];
	}
}

-(CheckPointerResult)checkPointer:(const void *)inPtr printObjectDescription:(BOOL)inPrintObjectDescription
{
	if(IsObjcObject(inPtr))
	{
		Class taggedPointerClass = NULL;
		if(IsObjcTaggedPointer(inPtr, &taggedPointerClass))
		{
			NSString *className = @"unknown";
			if(taggedPointerClass != nil)
			{
				className = NSStringFromClass(taggedPointerClass);
			}
			
			if([className containsString:@"NSTaggedPointerString"] ||
				[className containsString:@"__NSCFNumber"])
			{
				id theObject = (__bridge id)(void *)inPtr;
				if(inPrintObjectDescription)
				{
					NSString *logText = [NSString stringWithFormat:@"%p -> %@ tagged pointer: %@", inPtr, className, [theObject description]];
					[self logText:logText isError:NO];
				}
				else
				{
					NSString *logText = [NSString stringWithFormat:@"%p -> %@ tagged pointer", inPtr, className];
					[self logText:logText isError:NO];
				}
			}
			else
			{
				NSString *logText = [NSString stringWithFormat:@"%p -> %@ tagged pointer: (no description)", inPtr, className];
				[self logText:logText isError:NO];
			}
			
			return kCheckPointerResult_TaggedPointed;
		}
		else
		{
			id theObject = (__bridge id)(void *)inPtr;
			if([theObject respondsToSelector:@selector(description)])
			{
				if(inPrintObjectDescription)
				{
					NSString *logText = [NSString stringWithFormat:@"%p -> %@: %@", inPtr, NSStringFromClass([theObject class]), [theObject description]];
					[self logText:logText isError:NO];
				}
				else
				{
					NSString *logText = [NSString stringWithFormat:@"%p -> %@", inPtr, NSStringFromClass([theObject class])];
					[self logText:logText isError:NO];
				}
			}
			else if(theObject != nil)
			{
				NSString *logText = [NSString stringWithFormat:@"%p -> %@ (no description)", inPtr, NSStringFromClass([theObject class])];
				[self logText:logText isError:NO];
			}
			else
			{
				NSString *logText = [NSString stringWithFormat:@"%p -> (no description)", inPtr];
				[self logText:logText isError:NO];
			}
			
			return kCheckPointerResult_Object;
		}
	}
	else
	{
		NSString *logText = [NSString stringWithFormat:@"%p -> not an object", inPtr];
		[self logText:logText isError:NO];
		return kCheckPointerResult_NotAnObject;
	}
}

#pragma  mark - Logging

-(void)logText:(NSString *)inString isError:(BOOL)inIsError
{
	static NSDictionary *sErrorAttributes = nil;
	if(sErrorAttributes == nil)
	{
		sErrorAttributes = @{ NSForegroundColorAttributeName : [UIColor redColor] };
	}
	
	static UIFont *sBoldFont = nil;
	if(sBoldFont == nil)
	{
		sBoldFont = [UIFont boldSystemFontOfSize:12];
	}
	
	// Get the attributed string
	NSMutableAttributedString *textViewText = [self.textView.attributedText mutableCopy];
	
	// Append the text
	NSString *theString = [NSString stringWithFormat:@"%@\n", inString];
	NSAttributedString *appendText = nil;
	if(inIsError)
	{
		appendText = [[NSAttributedString alloc] initWithString:theString attributes:sErrorAttributes];
	}
	else
	{
		// Display the object description in bold
		NSRange objectDescriptionStartRange = [theString rangeOfString:@": "];
		if(objectDescriptionStartRange.location != NSNotFound)
		{
			NSUInteger boldStartLocation = objectDescriptionStartRange.location + 2;
			NSRange objectDescriptionRange = NSMakeRange(boldStartLocation, [theString length] - boldStartLocation - 1);
			NSMutableAttributedString *attributedString = [[NSMutableAttributedString alloc] initWithString:theString];
			[attributedString addAttribute:NSFontAttributeName value:sBoldFont range:objectDescriptionRange];
			appendText = attributedString;
		}
		else
		{
			appendText = [[NSAttributedString alloc] initWithString:theString];
		}
	}
	
	[textViewText appendAttributedString:appendText];
	
	// Set the new attributed string
	self.textView.attributedText = textViewText;
	
	// Scroll to bottom
	[self.textView scrollRangeToVisible:NSMakeRange([self.textView.text length], 0)];
	
	// Log to console too
	NSLog(@"%@", inString);
}

@end
